@@ -1,27 +1,56 @@
 """Interactive figures in the Jupyter notebook"""
 
-from base64 import b64encode
-import json
+import sys
+import types
+
+# In the case of a pyodide context (JupyterLite)
+# we mock Tornado, as it cannot be imported and would
+# lead to errors. Of course, any code using tornado cannot
+# work with this workaround, we prefer a half-working
+# solution than a non-working solution here.
+try:
+    import micropip  # noqa
+except ImportError:
+    sys.modules['tornado'] = types.ModuleType('tornadofake')
+
 import io
+import json
+from base64 import b64encode
 
-from IPython.display import display, HTML
-from IPython import get_ipython
-from IPython import version_info as ipython_version_info
-
-from ipywidgets import DOMWidget, widget_serialization
-from traitlets import (
-    Unicode, Bool, CInt, List, Instance, CaselessStrEnum, Enum,
-    default
-)
+try:
+    from collections.abc import Iterable
+except ImportError:
+    # Python 2.7
+    from collections import Iterable
 
 import matplotlib
-from matplotlib import rcParams, is_interactive
-from matplotlib.backends.backend_webagg_core import (FigureManagerWebAgg,
-                                                     FigureCanvasWebAggCore,
-                                                     NavigationToolbar2WebAgg,
-                                                     TimerTornado)
-from matplotlib.backend_bases import NavigationToolbar2, cursors, _Backend
+import numpy as np
+from IPython import get_ipython
+from IPython import version_info as ipython_version_info
+from IPython.display import HTML, display
+from ipython_genutils.py3compat import string_types
+from ipywidgets import DOMWidget, widget_serialization
+from matplotlib import is_interactive, rcParams
 from matplotlib._pylab_helpers import Gcf
+from matplotlib.backend_bases import NavigationToolbar2, _Backend, cursors
+from matplotlib.backends.backend_webagg_core import (
+    FigureCanvasWebAggCore,
+    FigureManagerWebAgg,
+    NavigationToolbar2WebAgg,
+    TimerTornado,
+)
+from PIL import Image
+from traitlets import (
+    Any,
+    Bool,
+    CaselessStrEnum,
+    CInt,
+    Enum,
+    Instance,
+    List,
+    Unicode,
+    default,
+)
 
 from ._version import js_semver
 
@@ -30,7 +59,7 @@ cursors_str = {
     cursors.POINTER: 'default',
     cursors.SELECT_REGION: 'crosshair',
     cursors.MOVE: 'move',
-    cursors.WAIT: 'wait'
+    cursors.WAIT: 'wait',
 }
 
 
@@ -44,11 +73,14 @@ def connection_info():
     result = []
     for manager in Gcf.get_all_fig_managers():
         fig = manager.canvas.figure
-        result.append('{0} - {1}'.format((fig.get_label() or
-                                          "Figure {}".format(manager.num)),
-                                         manager.web_sockets))
+        result.append(
+            '{} - {}'.format(
+                (fig.get_label() or f"Figure {manager.num}"),
+                manager.web_sockets,
+            )
+        )
     if not is_interactive():
-        result.append('Figures pending show: {0}'.format(len(Gcf._activeQue)))
+        result.append(f'Figures pending show: {len(Gcf._activeQue)}')
     return '\n'.join(result)
 
 
@@ -63,16 +95,17 @@ class Toolbar(DOMWidget, NavigationToolbar2WebAgg):
     _view_name = Unicode('ToolbarView').tag(sync=True)
 
     toolitems = List().tag(sync=True)
-    orientation = Enum(['horizontal', 'vertical'],
-                       default_value='vertical').tag(sync=True)
+    orientation = Enum(['horizontal', 'vertical'], default_value='vertical').tag(
+        sync=True
+    )
     button_style = CaselessStrEnum(
         values=['primary', 'success', 'info', 'warning', 'danger', ''],
         default_value='',
-        help="""Use a predefined styling for the button.""").tag(sync=True)
+        help="""Use a predefined styling for the button.""",
+    ).tag(sync=True)
     collapsed = Bool(True).tag(sync=True)
 
-    _current_action = Enum(values=['pan', 'zoom', ''],
-                           default_value='').tag(sync=True)
+    _current_action = Enum(values=['pan', 'zoom', ''], default_value='').tag(sync=True)
 
     def __init__(self, canvas, *args, **kwargs):
         DOMWidget.__init__(self, *args, **kwargs)
@@ -83,8 +116,15 @@ class Toolbar(DOMWidget, NavigationToolbar2WebAgg):
     def export(self):
         buf = io.BytesIO()
         self.canvas.figure.savefig(buf, format='png', dpi='figure')
-        data = "<img src='data:image/png;base64,{0}'/>"
-        data = data.format(b64encode(buf.getvalue()).decode('utf-8'))
+        # Figure width in pixels
+        pwidth = self.canvas.figure.get_figwidth() * self.canvas.figure.get_dpi()
+        # Scale size to match widget on HiDPI monitors.
+        if hasattr(self.canvas, 'device_pixel_ratio'):  # Matplotlib 3.5+
+            width = pwidth / self.canvas.device_pixel_ratio
+        else:
+            width = pwidth / self.canvas._dpi_ratio
+        data = "<img src='data:image/png;base64,{0}' width={1}/>"
+        data = data.format(b64encode(buf.getvalue()).decode('utf-8'), width)
         display(HTML(data))
 
     @default('toolitems')
@@ -96,18 +136,18 @@ class Toolbar(DOMWidget, NavigationToolbar2WebAgg):
             'zoom_to_rect': 'square-o',
             'move': 'arrows',
             'download': 'floppy-o',
-            'export': 'file-picture-o'
+            'export': 'file-picture-o',
         }
 
-        download_item = ('Download', 'Download plot', 'download',
-                         'save_figure')
+        download_item = ('Download', 'Download plot', 'download', 'save_figure')
 
-        toolitems = (NavigationToolbar2.toolitems + (download_item,))
+        toolitems = NavigationToolbar2.toolitems + (download_item,)
 
-        return [(text, tooltip, icons[icon_name], method_name)
-                for text, tooltip, icon_name, method_name
-                in toolitems
-                if icon_name in icons]
+        return [
+            (text, tooltip, icons[icon_name], method_name)
+            for text, tooltip, icon_name, method_name in toolitems
+            if icon_name in icons
+        ]
 
 
 class Canvas(DOMWidget, FigureCanvasWebAggCore):
@@ -120,11 +160,11 @@ class Canvas(DOMWidget, FigureCanvasWebAggCore):
     _view_module_version = Unicode(js_semver).tag(sync=True)
     _view_name = Unicode('MPLCanvasView').tag(sync=True)
 
-    toolbar = Instance(Toolbar,
-                       allow_none=True).tag(sync=True, **widget_serialization)
+    toolbar = Instance(Toolbar, allow_none=True).tag(sync=True, **widget_serialization)
     toolbar_visible = Bool(True).tag(sync=True)
-    toolbar_position = Enum(['top', 'bottom', 'left', 'right'],
-                            default_value='left').tag(sync=True)
+    toolbar_position = Enum(
+        ['top', 'bottom', 'left', 'right'], default_value='left'
+    ).tag(sync=True)
 
     header_visible = Bool(True).tag(sync=True)
     footer_visible = Bool(True).tag(sync=True)
@@ -132,6 +172,14 @@ class Canvas(DOMWidget, FigureCanvasWebAggCore):
     resizable = Bool(True).tag(sync=True)
     capture_scroll = Bool(False).tag(sync=True)
 
+    # This is a very special widget trait:
+    # We set "sync=True" because we want ipywidgets to consider this
+    # as part of the widget state, but we overwrite send_state so that
+    # it never sync the value with the front-end, the front-end keeps its
+    # own value.
+    # This will still be used by ipywidgets in the case of embedding.
+    _data_url = Any(None).tag(sync=True)
+
     _width = CInt().tag(sync=True)
     _height = CInt().tag(sync=True)
 
@@ -162,12 +210,34 @@ class Canvas(DOMWidget, FigureCanvasWebAggCore):
 
         self.on_msg(self._handle_message)
 
+        # This will stay True for cases where there is no
+        # front-end (e.g. nbconvert --execute)
+        self.syncing_data_url = True
+
+    # Overwrite ipywidgets's send_state so we don't sync the data_url
+    def send_state(self, key=None):
+        if key is None:
+            keys = self.keys
+        elif isinstance(key, string_types):
+            keys = [key]
+        elif isinstance(key, Iterable):
+            keys = key
+
+        if not self.syncing_data_url:
+            keys = [k for k in keys if k != '_data_url']
+
+        DOMWidget.send_state(self, key=keys)
+
     def _handle_message(self, object, content, buffers):
         # Every content has a "type".
         if content['type'] == 'closing':
             self._closed = True
 
         elif content['type'] == 'initialized':
+            # We stop syncing data url, the front-end is there and
+            # ready to receive diffs
+            self.syncing_data_url = False
+
             _, _, w, h = self.figure.bbox.bounds
             self.manager.resize(w, h)
 
@@ -182,9 +252,7 @@ class Canvas(DOMWidget, FigureCanvasWebAggCore):
         # Change in the widget state?
         if content['type'] == 'cursor':
             cursor = content['cursor']
-            self._cursor = (
-                cursors_str[cursor] if cursor in cursors_str else cursor
-            )
+            self._cursor = cursors_str[cursor] if cursor in cursors_str else cursor
 
         elif content['type'] == 'message':
             self._message = content['message']
@@ -206,6 +274,19 @@ class Canvas(DOMWidget, FigureCanvasWebAggCore):
             self.send({'data': json.dumps(content)})
 
     def send_binary(self, data):
+        # TODO we should maybe rework the FigureCanvasWebAggCore implementation
+        # so that it has a "refresh" method that we can overwrite
+
+        # Update _data_url
+        if self.syncing_data_url:
+            data = self._last_buff.view(dtype=np.uint8).reshape(
+                (*self._last_buff.shape, 4)
+            )
+            with io.BytesIO() as png:
+                Image.fromarray(data).save(png, format="png")
+                self._data_url = b64encode(png.getvalue()).decode('utf-8')
+
+        # Actually send the data
         self.send({'data': '{"type": "binary"}'}, buffers=[data])
 
     def new_timer(self, *args, **kwargs):
@@ -221,16 +302,33 @@ class Canvas(DOMWidget, FigureCanvasWebAggCore):
 
         buf = io.BytesIO()
         self.figure.savefig(buf, format='png', dpi='figure')
-        data_url = b64encode(buf.getvalue()).decode('utf-8')
+        self._data_url = b64encode(buf.getvalue()).decode('utf-8')
+        # Figure width in pixels
+        pwidth = self.figure.get_figwidth() * self.figure.get_dpi()
+        # Scale size to match widget on HiDPI monitors.
+        if hasattr(self, 'device_pixel_ratio'):  # Matplotlib 3.5+
+            width = pwidth / self.device_pixel_ratio
+        else:
+            width = pwidth / self._dpi_ratio
+        html = """
+            <div style="display: inline-block;">
+                <div class="jupyter-widgets widget-label" style="text-align: center;">
+                    {}
+                </div>
+                <img src='data:image/png;base64,{}' width={}/>
+            </div>
+        """.format(
+            self._figure_label, self._data_url, width
+        )
 
         data = {
             'text/plain': plaintext,
-            'image/png': data_url,
+            'text/html': html,
             'application/vnd.jupyter.widget-view+json': {
                 'version_major': 2,
                 'version_minor': 0,
-                'model_id': self._model_id
-            }
+                'model_id': self._model_id,
+            },
         }
 
         return data
@@ -249,54 +347,56 @@ class Canvas(DOMWidget, FigureCanvasWebAggCore):
     if matplotlib.__version__ < '3.4':
         # backport the Python side changes to match the js changes
         def _handle_key(self, event):
-            _SPECIAL_KEYS_LUT = {'Alt': 'alt',
-                                 'AltGraph': 'alt',
-                                 'CapsLock': 'caps_lock',
-                                 'Control': 'control',
-                                 'Meta': 'meta',
-                                 'NumLock': 'num_lock',
-                                 'ScrollLock': 'scroll_lock',
-                                 'Shift': 'shift',
-                                 'Super': 'super',
-                                 'Enter': 'enter',
-                                 'Tab': 'tab',
-                                 'ArrowDown': 'down',
-                                 'ArrowLeft': 'left',
-                                 'ArrowRight': 'right',
-                                 'ArrowUp': 'up',
-                                 'End': 'end',
-                                 'Home': 'home',
-                                 'PageDown': 'pagedown',
-                                 'PageUp': 'pageup',
-                                 'Backspace': 'backspace',
-                                 'Delete': 'delete',
-                                 'Insert': 'insert',
-                                 'Escape': 'escape',
-                                 'Pause': 'pause',
-                                 'Select': 'select',
-                                 'Dead': 'dead',
-                                 'F1': 'f1',
-                                 'F2': 'f2',
-                                 'F3': 'f3',
-                                 'F4': 'f4',
-                                 'F5': 'f5',
-                                 'F6': 'f6',
-                                 'F7': 'f7',
-                                 'F8': 'f8',
-                                 'F9': 'f9',
-                                 'F10': 'f10',
-                                 'F11': 'f11',
-                                 'F12': 'f12'}
+            _SPECIAL_KEYS_LUT = {
+                'Alt': 'alt',
+                'AltGraph': 'alt',
+                'CapsLock': 'caps_lock',
+                'Control': 'control',
+                'Meta': 'meta',
+                'NumLock': 'num_lock',
+                'ScrollLock': 'scroll_lock',
+                'Shift': 'shift',
+                'Super': 'super',
+                'Enter': 'enter',
+                'Tab': 'tab',
+                'ArrowDown': 'down',
+                'ArrowLeft': 'left',
+                'ArrowRight': 'right',
+                'ArrowUp': 'up',
+                'End': 'end',
+                'Home': 'home',
+                'PageDown': 'pagedown',
+                'PageUp': 'pageup',
+                'Backspace': 'backspace',
+                'Delete': 'delete',
+                'Insert': 'insert',
+                'Escape': 'escape',
+                'Pause': 'pause',
+                'Select': 'select',
+                'Dead': 'dead',
+                'F1': 'f1',
+                'F2': 'f2',
+                'F3': 'f3',
+                'F4': 'f4',
+                'F5': 'f5',
+                'F6': 'f6',
+                'F7': 'f7',
+                'F8': 'f8',
+                'F9': 'f9',
+                'F10': 'f10',
+                'F11': 'f11',
+                'F12': 'f12',
+            }
 
             def handle_key(key):
                 """Handle key values"""
-                value = key[key.index('k') + 1:]
+                value = key[key.index('k') + 1 :]
                 if 'shift+' in key:
                     if len(value) == 1:
                         key = key.replace('shift+', '')
                 if value in _SPECIAL_KEYS_LUT:
                     value = _SPECIAL_KEYS_LUT[value]
-                key = key[:key.index('k')] + value
+                key = key[: key.index('k')] + value
                 return key
 
             key = handle_key(event['key'])
@@ -306,6 +406,7 @@ class Canvas(DOMWidget, FigureCanvasWebAggCore):
                 self.key_press_event(key, guiEvent=guiEvent)
             elif e_type == 'key_release':
                 self.key_release_event(key, guiEvent=guiEvent)
+
         handle_key_press = handle_key_release = _handle_key
 
 
@@ -413,12 +514,9 @@ def flush_figures():
 
         try:
             # exclude any figures that were closed:
-            active = set([
-                fm.canvas.figure for fm in Gcf.get_all_fig_managers()
-            ])
+            active = {fm.canvas.figure for fm in Gcf.get_all_fig_managers()}
 
-            for fig in [
-                    fig for fig in _Backend_ipympl._to_show if fig in active]:
+            for fig in [fig for fig in _Backend_ipympl._to_show if fig in active]:
                 # display(fig.canvas, metadata=_fetch_figure_metadata(fig))
                 display(fig.canvas)
         finally:
@@ -428,4 +526,5 @@ def flush_figures():
 
 
 ip = get_ipython()
-ip.events.register('post_execute', flush_figures)
+if ip is not None:
+    ip.events.register('post_execute', flush_figures)
